/*************************************************************************************************
 * Perl binding
 *                                                               Copyright (C) 2009-2010 FAL Labs
 * This file is part of Kyoto Cabinet.
 * This program is free software: you can redistribute it and/or modify it under the terms of
 * the GNU General Public License as published by the Free Software Foundation, either version
 * 3 of the License, or any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License along with this program.
 * If not, see <http://www.gnu.org/licenses/>.
 *************************************************************************************************/


#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include <kclangc.h>


/**
 * Arguments for visitor functions.
 */
typedef struct {
  SV* svvisitor;
  int32_t writable;
  char* rbuf;
} VISARG;



/* global variables */
SV* sv_vis_nop;
SV* sv_vis_remove;


/* function prototypes */
static int64_t svatoi(SV* sv);
static double svatof(SV* sv);
static const char* softvisitfull(const char* kbuf, size_t ksiz,
                                 const char* vbuf, size_t vsiz, size_t* sp, void* opq);
static const char* softvisitempty(const char* kbuf, size_t ksiz, size_t* sp, void* opq);
static int32_t softfileproc(const char* path, int64_t count, int64_t size, void* opq);


/**
 * Convert a numeric parameter to an integer.
 */
static int64_t svatoi(SV* sv) {
  if (SvIOK(sv)) {
    return SvIV(sv);
  } else if (SvNOK(sv)) {
    double dnum = SvNV(sv);
    if (kcchknan(dnum)) {
      return INT64_MIN;
    } else if (kcchkinf(dnum)) {
      return dnum < 0 ? INT64_MIN : INT64_MAX;
    }
    return dnum;
  } else {
    size_t nsiz;
    const char* nbuf = SvPV(sv, nsiz);
    double dnum = kcatof(nbuf);
    if (kcchknan(dnum)) {
      return INT64_MIN;
    } else if (kcchkinf(dnum)) {
      return dnum < 0 ? INT64_MIN : INT64_MAX;
    }
    return dnum;
  }
  return 0;
}


/**
 * Convert a numeric parameter to a real number.
 */
static double svatof(SV* sv) {
  if (SvIOK(sv)) {
    return SvIV(sv);
  } else if (SvNOK(sv)) {
    return SvNV(sv);
  } else {
    size_t nsiz;
    const char* nbuf = SvPV(sv, nsiz);
    return kcatof(nbuf);
  }
  return 0;
}


/**
 * Call back function to visit a full record.
 */
static const char* softvisitfull(const char* kbuf, size_t ksiz,
                                 const char* vbuf, size_t vsiz, size_t* rsp, void* opq) {
  VISARG* arg = (VISARG*)opq;
  SV* svvisitor = arg->svvisitor;
  if (arg->rbuf) kcfree(arg->rbuf);
  arg->rbuf = NULL;
  const char* rv = KCVISNOP;
  if (SvROK(svvisitor)) {
    svtype t = SvTYPE(SvRV(svvisitor));
    if (t == SVt_PVCV) {
      dSP;
      ENTER;
      SAVETMPS;
      PUSHMARK(SP);
      XPUSHs(sv_2mortal(newSVpvn(kbuf, ksiz)));
      XPUSHs(sv_2mortal(newSVpvn(vbuf, vsiz)));
      PUTBACK;
      int32_t res = call_sv(svvisitor, G_SCALAR | G_EVAL);
      SPAGAIN;
      if (res == 1) {
        SV* svrv = POPs;
        if (SvOK(svrv)) {
           if (SvROK(svrv)) {
             if (SvRV(svrv) == sv_vis_remove) {
               rv = KCVISREMOVE;
             } else if (SvRV(svrv) != sv_vis_nop) {
               svrv = SvRV(svrv);
               size_t rsiz;
               const char* rbuf = SvPV(svrv, rsiz);
               arg->rbuf = kcmalloc(rsiz + 1);
               memcpy(arg->rbuf, rbuf, rsiz);
               rv = arg->rbuf;
               *rsp = rsiz;
             }
           } else {
             size_t rsiz;
             const char* rbuf = SvPV(svrv, rsiz);
             arg->rbuf = kcmalloc(rsiz + 1);
             memcpy(arg->rbuf, rbuf, rsiz);
             rv = arg->rbuf;
             *rsp = rsiz;
           }
        }
      } else {
        croak("call_sv failed\n");
      }
      PUTBACK;
      FREETMPS;
      LEAVE;
    } else {
      dSP;
      ENTER;
      SAVETMPS;
      PUSHMARK(SP);
      XPUSHs(svvisitor);
      XPUSHs(sv_2mortal(newSVpvn(kbuf, ksiz)));
      XPUSHs(sv_2mortal(newSVpvn(vbuf, vsiz)));
      PUTBACK;
      int32_t res = call_method("visit_full", G_SCALAR | G_EVAL);
      SPAGAIN;
      if (res == 1) {
        SV* svrv = POPs;
        if (SvOK(svrv)) {
           if (SvROK(svrv)) {
             if (SvRV(svrv) == sv_vis_remove) {
               rv = KCVISREMOVE;
             } else if (SvRV(svrv) != sv_vis_nop) {
               svrv = SvRV(svrv);
               size_t rsiz;
               const char* rbuf = SvPV(svrv, rsiz);
               arg->rbuf = kcmalloc(rsiz + 1);
               memcpy(arg->rbuf, rbuf, rsiz);
               rv = arg->rbuf;
               *rsp = rsiz;
             }
           } else {
             size_t rsiz;
             const char* rbuf = SvPV(svrv, rsiz);
             arg->rbuf = kcmalloc(rsiz + 1);
             memcpy(arg->rbuf, rbuf, rsiz);
             rv = arg->rbuf;
             *rsp = rsiz;
           }
        }
      } else {
        croak("call_sv failed\n");
      }
      PUTBACK;
      FREETMPS;
      LEAVE;
    }
  } else if (SvPOK(svvisitor)) {
    dSP;
    ENTER;
    SAVETMPS;
    PUSHMARK(SP);
    XPUSHs(sv_2mortal(newSVpvn(kbuf, ksiz)));
    XPUSHs(sv_2mortal(newSVpvn(vbuf, vsiz)));
    PUTBACK;
    int32_t res = call_sv(svvisitor, G_SCALAR | G_EVAL);
    SPAGAIN;
    if (res == 1) {
      SV* svrv = POPs;
      if (SvOK(svrv)) {
        if (SvROK(svrv)) {
          if (SvRV(svrv) == sv_vis_remove) {
            rv = KCVISREMOVE;
          } else if (SvRV(svrv) != sv_vis_nop) {
            svrv = SvRV(svrv);
            size_t rsiz;
            const char* rbuf = SvPV(svrv, rsiz);
            arg->rbuf = kcmalloc(rsiz + 1);
            memcpy(arg->rbuf, rbuf, rsiz);
            rv = arg->rbuf;
            *rsp = rsiz;
          }
        } else {
          size_t rsiz;
          const char* rbuf = SvPV(svrv, rsiz);
          arg->rbuf = kcmalloc(rsiz + 1);
          memcpy(arg->rbuf, rbuf, rsiz);
          rv = arg->rbuf;
          *rsp = rsiz;
        }
      }
    } else {
      croak("call_sv failed\n");
    }
    PUTBACK;
    FREETMPS;
    LEAVE;
  }
  return rv;
}


/**
 * Call back function to visit an empty record.
 */
static const char* softvisitempty(const char* kbuf, size_t ksiz, size_t* rsp, void* opq) {
  VISARG* arg = (VISARG*)opq;
  SV* svvisitor = arg->svvisitor;
  if (arg->rbuf) kcfree(arg->rbuf);
  arg->rbuf = NULL;
  const char* rv = KCVISNOP;
  if (SvROK(svvisitor)) {
    svtype t = SvTYPE(SvRV(svvisitor));
    if (t == SVt_PVCV) {
      dSP;
      ENTER;
      SAVETMPS;
      PUSHMARK(SP);
      XPUSHs(sv_2mortal(newSVpvn(kbuf, ksiz)));
      PUTBACK;
      int32_t res = call_sv(svvisitor, G_SCALAR | G_EVAL);
      SPAGAIN;
      if (res == 1) {
        SV* svrv = POPs;
        if (SvOK(svrv)) {
           if (SvROK(svrv)) {
             if (SvRV(svrv) == sv_vis_remove) {
               rv = KCVISREMOVE;
             } else if (SvRV(svrv) != sv_vis_nop) {
               svrv = SvRV(svrv);
               size_t rsiz;
               const char* rbuf = SvPV(svrv, rsiz);
               arg->rbuf = kcmalloc(rsiz + 1);
               memcpy(arg->rbuf, rbuf, rsiz);
               rv = arg->rbuf;
               *rsp = rsiz;
             }
           } else {
             size_t rsiz;
             const char* rbuf = SvPV(svrv, rsiz);
             arg->rbuf = kcmalloc(rsiz + 1);
             memcpy(arg->rbuf, rbuf, rsiz);
             rv = arg->rbuf;
             *rsp = rsiz;
           }
        }
      } else {
        croak("call_sv failed\n");
      }
      PUTBACK;
      FREETMPS;
      LEAVE;
    } else {
      dSP;
      ENTER;
      SAVETMPS;
      PUSHMARK(SP);
      XPUSHs(svvisitor);
      XPUSHs(sv_2mortal(newSVpvn(kbuf, ksiz)));
      PUTBACK;
      int32_t res = call_method("visit_empty", G_SCALAR | G_EVAL);
      SPAGAIN;
      if (res == 1) {
        SV* svrv = POPs;
        if (SvOK(svrv)) {
           if (SvROK(svrv)) {
             if (SvRV(svrv) == sv_vis_remove) {
               rv = KCVISREMOVE;
             } else if (SvRV(svrv) != sv_vis_nop) {
               svrv = SvRV(svrv);
               size_t rsiz;
               const char* rbuf = SvPV(svrv, rsiz);
               arg->rbuf = kcmalloc(rsiz + 1);
               memcpy(arg->rbuf, rbuf, rsiz);
               rv = arg->rbuf;
               *rsp = rsiz;
             }
           } else {
             size_t rsiz;
             const char* rbuf = SvPV(svrv, rsiz);
             arg->rbuf = kcmalloc(rsiz + 1);
             memcpy(arg->rbuf, rbuf, rsiz);
             rv = arg->rbuf;
             *rsp = rsiz;
           }
        }
      } else {
        croak("call_sv failed\n");
      }
      PUTBACK;
      FREETMPS;
      LEAVE;
    }
  } else if (SvPOK(svvisitor)) {
    dSP;
    ENTER;
    SAVETMPS;
    PUSHMARK(SP);
    XPUSHs(sv_2mortal(newSVpvn(kbuf, ksiz)));
    PUTBACK;
    int32_t res = call_sv(svvisitor, G_SCALAR | G_EVAL);
    SPAGAIN;
    if (res == 1) {
      SV* svrv = POPs;
      if (SvOK(svrv)) {
        if (SvROK(svrv)) {
          if (SvRV(svrv) == sv_vis_remove) {
            rv = KCVISREMOVE;
          } else if (SvRV(svrv) != sv_vis_nop) {
            svrv = SvRV(svrv);
            size_t rsiz;
            const char* rbuf = SvPV(svrv, rsiz);
            arg->rbuf = kcmalloc(rsiz + 1);
            memcpy(arg->rbuf, rbuf, rsiz);
            rv = arg->rbuf;
            *rsp = rsiz;
          }
        } else {
          size_t rsiz;
          const char* rbuf = SvPV(svrv, rsiz);
          arg->rbuf = kcmalloc(rsiz + 1);
          memcpy(arg->rbuf, rbuf, rsiz);
          rv = arg->rbuf;
          *rsp = rsiz;
        }
      }
    } else {
      croak("call_sv failed\n");
    }
    PUTBACK;
    FREETMPS;
    LEAVE;
  }
  return rv;
}


/**
 * Call back function to process the database file.
 */
static int32_t softfileproc(const char* path, int64_t count, int64_t size, void* opq) {
  SV* svproc = (SV*)opq;
  int32_t rv = 1;
  if (SvROK(svproc)) {
    svtype t = SvTYPE(SvRV(svproc));
    if (t == SVt_PVCV) {
      dSP;
      ENTER;
      SAVETMPS;
      PUSHMARK(SP);
      XPUSHs(sv_2mortal(newSVpv(path, 0)));
      XPUSHs(sv_2mortal(newSViv(count)));
      XPUSHs(sv_2mortal(newSViv(size)));
      PUTBACK;
      int32_t res = call_sv(svproc, G_SCALAR | G_EVAL);
      SPAGAIN;
      if (res == 1) {
        SV* svrv = POPs;
        rv = SvTRUE(svrv);
      } else {
        croak("call_sv failed\n");
      }
      PUTBACK;
      FREETMPS;
      LEAVE;
    } else {
      dSP;
      ENTER;
      SAVETMPS;
      PUSHMARK(SP);
      XPUSHs(svproc);
      XPUSHs(sv_2mortal(newSVpv(path, 0)));
      XPUSHs(sv_2mortal(newSViv(count)));
      XPUSHs(sv_2mortal(newSViv(size)));
      PUTBACK;
      int32_t res = call_method("process", G_SCALAR | G_EVAL);
      SPAGAIN;
      if (res == 1) {
        SV* svrv = POPs;
        rv = SvTRUE(svrv);
      } else {
        croak("call_sv failed\n");
      }
      PUTBACK;
      FREETMPS;
      LEAVE;
    }
  } else if (SvPOK(svproc)) {
    dSP;
    ENTER;
    SAVETMPS;
    PUSHMARK(SP);
    XPUSHs(sv_2mortal(newSVpv(path, 0)));
    XPUSHs(sv_2mortal(newSViv(count)));
    XPUSHs(sv_2mortal(newSViv(size)));
    PUTBACK;
    int32_t res = call_sv(svproc, G_SCALAR | G_EVAL);
    SPAGAIN;
    if (res == 1) {
      SV* svrv = POPs;
      rv = SvTRUE(svrv);
    } else {
      croak("call_sv failed\n");
    }
    PUTBACK;
    FREETMPS;
    LEAVE;
  }
  return rv;
}



MODULE = KyotoCabinet		PACKAGE = KyotoCabinet
PROTOTYPES: DISABLE


const char*
VERSION()
PPCODE:
	XPUSHs(sv_2mortal(newSVpv(KCVERSION, 0)));


void
atoi(SV* svstr)
PPCODE:
	size_t size;
	const char* buf = SvPV(svstr, size);
	XPUSHs(sv_2mortal(newSViv(kcatoi(buf))));
	XSRETURN(1);


void
atoix(SV* svstr)
PPCODE:
	size_t size;
	const char* buf = SvPV(svstr, size);
	XPUSHs(sv_2mortal(newSViv(kcatoix(buf))));
	XSRETURN(1);


void
atof(SV* svstr)
PPCODE:
	size_t size;
	const char* buf = SvPV(svstr, size);
	XPUSHs(sv_2mortal(newSVnv(kcatof(buf))));
	XSRETURN(1);


void
hash_murmur(SV* svstr)
PPCODE:
	size_t size;
	const char* buf = SvPV(svstr, size);
	XPUSHs(sv_2mortal(newSViv(kchashmurmur(buf, size))));
	XSRETURN(1);


void
hash_fnv(SV* svstr)
PPCODE:
	size_t size;
	const char* buf = SvPV(svstr, size);
	XPUSHs(sv_2mortal(newSViv(kchashfnv(buf, size))));
	XSRETURN(1);


void
levdist(SV* sva, SV* svb, ...)
PPCODE:
	size_t asiz;
	const char* abuf = SvPV(sva, asiz);
	size_t bsiz;
	const char* bbuf = SvPV(svb, bsiz);
	int32_t utf = items > 2 ? SvTRUE(ST(2)) : 0;
	XPUSHs(sv_2mortal(newSViv(kclevdist(abuf, asiz, bbuf, bsiz, utf))));
	XSRETURN(1);



MODULE = KyotoCabinet		PACKAGE = KyotoCabinet::Error
PROTOTYPES: DISABLE


void
SUCCESS(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCESUCCESS)));
	XSRETURN(1);


void
NOIMPL(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCENOIMPL)));
	XSRETURN(1);


void
INVALID(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCEINVALID)));
	XSRETURN(1);


void
NOREPOS(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCENOREPOS)));
	XSRETURN(1);


void
NOPERM(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCENOPERM)));
	XSRETURN(1);


void
BROKEN(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCEBROKEN)));
	XSRETURN(1);


void
DUPREC(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCEDUPREC)));
	XSRETURN(1);


void
NOREC(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCENOREC)));
	XSRETURN(1);


void
LOGIC(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCELOGIC)));
	XSRETURN(1);


void
SYSTEM(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCESYSTEM)));
	XSRETURN(1);


void
MISC(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCEMISC)));
	XSRETURN(1);


void
err_codename(IV code)
PPCODE:
	const char* name = kcecodename(code);
	XPUSHs(sv_2mortal(newSVpv(name, 0)));
	XSRETURN(1);




MODULE = KyotoCabinet		PACKAGE = KyotoCabinet::Visitor
PROTOTYPES: DISABLE


BOOT:
	sv_vis_nop = newSVpv("[NOP]", 0);
	sv_vis_remove = newSVpv("[REMOVE]", 0);


void
NOP(...)
PPCODE:
	XPUSHs(sv_2mortal(newRV_inc(sv_vis_nop)));
	XSRETURN(1);


void
REMOVE(...)
PPCODE:
	XPUSHs(sv_2mortal(newRV_inc(sv_vis_remove)));
	XSRETURN(1);



MODULE = KyotoCabinet		PACKAGE = KyotoCabinet::Cursor
PROTOTYPES: DISABLE


void
cur_new(IV ivdb)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	KCCUR* cur = kcdbcursor(db);
	XPUSHs(sv_2mortal(newSViv((IV)cur)));
	XSRETURN(1);


void
cur_delete(IV ivcur)
PPCODE:
	if (!ivcur) XSRETURN_UNDEF;
	KCCUR* cur = (KCCUR*)ivcur;
	kccurdel(cur);
	XSRETURN(0);


void
cur_disable(IV ivcur)
PPCODE:
	if (!ivcur) XSRETURN_UNDEF;
	KCCUR* cur = (KCCUR*)ivcur;
	kccurdel(cur);
	XSRETURN(0);


void
cur_accept(IV ivcur, SV* svvisitor, SV* svwritable, SV* svstep)
PPCODE:
	if (!ivcur) XSRETURN_UNDEF;
	KCCUR* cur = (KCCUR*)ivcur;
	int32_t writable = SvTRUE(svwritable);
	int32_t step = SvTRUE(svstep);
	VISARG arg;
	memset(&arg, 0, sizeof(arg));
	arg.svvisitor = svvisitor;
	arg.writable = writable;
	arg.rbuf = NULL;
	int32_t rv = kccuraccept(cur, softvisitfull, &arg, writable, step);
	if (arg.rbuf) kcfree(arg.rbuf);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
cur_set_value(IV ivcur, SV* svvalue, SV* svstep)
PPCODE:
	if (!ivcur) XSRETURN_UNDEF;
	KCCUR* cur = (KCCUR*)ivcur;
	size_t vsiz;
	const char* vbuf = SvPV(svvalue, vsiz);
	int32_t step = SvTRUE(svstep);
	int32_t rv = kccursetvalue(cur, vbuf, vsiz, step);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
cur_remove(IV ivcur)
PPCODE:
	if (!ivcur) XSRETURN_UNDEF;
	KCCUR* cur = (KCCUR*)ivcur;
	int32_t rv = kccurremove(cur);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
cur_get_key(IV ivcur, SV* svstep)
PPCODE:
	if (!ivcur) XSRETURN_UNDEF;
	KCCUR* cur = (KCCUR*)ivcur;
	int32_t step = SvTRUE(svstep);
	size_t ksiz;
	char* kbuf = kccurgetkey(cur, &ksiz, step);
	if (kbuf) {
	  XPUSHs(sv_2mortal(newSVpvn(kbuf, ksiz)));
	  kcfree(kbuf);
	} else {
	  XPUSHs(&PL_sv_undef);
	}
	XSRETURN(1);


void
cur_get_value(IV ivcur, SV* svstep)
PPCODE:
	if (!ivcur) XSRETURN_UNDEF;
	KCCUR* cur = (KCCUR*)ivcur;
	int32_t step = SvTRUE(svstep);
	size_t vsiz;
	char* vbuf = kccurgetvalue(cur, &vsiz, step);
	if (vbuf) {
	  XPUSHs(sv_2mortal(newSVpvn(vbuf, vsiz)));
	  kcfree(vbuf);
	} else {
	  XPUSHs(&PL_sv_undef);
	}
	XSRETURN(1);


void
cur_get(IV ivcur, SV* svstep)
PPCODE:
	if (!ivcur) XSRETURN_UNDEF;
	KCCUR* cur = (KCCUR*)ivcur;
	int32_t step = SvTRUE(svstep);
	size_t ksiz, vsiz;
	const char* vbuf;
	char* kbuf = kccurget(cur, &ksiz, &vbuf, &vsiz, step);
	if (!kbuf) XSRETURN(0);
	XPUSHs(sv_2mortal(newSVpvn(kbuf, ksiz)));
	XPUSHs(sv_2mortal(newSVpvn(vbuf, vsiz)));
	kcfree(kbuf);
	XSRETURN(2);


void
cur_seize(IV ivcur)
PPCODE:
	if (!ivcur) XSRETURN_UNDEF;
	KCCUR* cur = (KCCUR*)ivcur;
	size_t ksiz, vsiz;
	const char* vbuf;
	char* kbuf = kccurseize(cur, &ksiz, &vbuf, &vsiz);
	if (!kbuf) XSRETURN(0);
	XPUSHs(sv_2mortal(newSVpvn(kbuf, ksiz)));
	XPUSHs(sv_2mortal(newSVpvn(vbuf, vsiz)));
	kcfree(kbuf);
	XSRETURN(2);


void
cur_jump(IV ivcur, SV* svkey)
PPCODE:
	if (!ivcur) XSRETURN_UNDEF;
	KCCUR* cur = (KCCUR*)ivcur;
	int32_t rv;
	if (SvOK(svkey)) {
	  size_t ksiz;
	  const char* kbuf = SvPV(svkey, ksiz);
	  rv = kccurjumpkey(cur, kbuf, ksiz);
	} else {
	  rv = kccurjump(cur);
	}
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
cur_jump_back(IV ivcur, SV* svkey)
PPCODE:
	if (!ivcur) XSRETURN_UNDEF;
	KCCUR* cur = (KCCUR*)ivcur;
	int32_t rv;
	if (SvOK(svkey)) {
	  size_t ksiz;
	  const char* kbuf = SvPV(svkey, ksiz);
	  rv = kccurjumpbackkey(cur, kbuf, ksiz);
	} else {
	  rv = kccurjumpback(cur);
	}
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
cur_step(IV ivcur)
PPCODE:
	if (!ivcur) XSRETURN_UNDEF;
	KCCUR* cur = (KCCUR*)ivcur;
	int32_t rv = kccurstep(cur);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
cur_step_back(IV ivcur)
PPCODE:
	if (!ivcur) XSRETURN_UNDEF;
	KCCUR* cur = (KCCUR*)ivcur;
	int32_t rv = kccurstepback(cur);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;



MODULE = KyotoCabinet		PACKAGE = KyotoCabinet::DB
PROTOTYPES: DISABLE


void
OREADER(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCOREADER)));
	XSRETURN(1);


void
OWRITER(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCOWRITER)));
	XSRETURN(1);


void
OCREATE(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCOCREATE)));
	XSRETURN(1);


void
OTRUNCATE(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCOTRUNCATE)));
	XSRETURN(1);


void
OAUTOTRAN(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCOAUTOTRAN)));
	XSRETURN(1);


void
OAUTOSYNC(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCOAUTOSYNC)));
	XSRETURN(1);


void
ONOLOCK(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCONOLOCK)));
	XSRETURN(1);


void
OTRYLOCK(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCOTRYLOCK)));
	XSRETURN(1);


void
ONOREPAIR(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCONOREPAIR)));
	XSRETURN(1);


void
MSET(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCMSET)));
	XSRETURN(1);


void
MADD(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCMADD)));
	XSRETURN(1);


void
MREPLACE(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCMREPLACE)));
	XSRETURN(1);


void
MAPPEND(...)
PPCODE:
	XPUSHs(sv_2mortal(newSViv(KCMAPPEND)));
	XSRETURN(1);


void
db_new()
PPCODE:
	KCDB* db = kcdbnew();
	XPUSHs(sv_2mortal(newSViv((IV)db)));
	XSRETURN(1);


void
db_delete(IV ivdb)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	kcdbdel(db);
	XSRETURN(0);


void
db_error(IV ivdb)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	int32_t code = kcdbecode(db);
	const char* message = kcdbemsg(db);
	XPUSHs(sv_2mortal(newSViv(code)));
	XPUSHs(sv_2mortal(newSVpv(message, 0)));
	XSRETURN(2);


void
db_open(IV ivdb, SV* svpath, IV mode)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t psiz;
	const char* path = SvPV(svpath, psiz);
	int32_t rv = kcdbopen(db, path, mode);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_accept(IV ivdb, SV* svkey, SV* svvisitor, SV* svwritable)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t ksiz;
	const char* kbuf = SvPV(svkey, ksiz);
	int32_t writable = SvTRUE(svwritable);
	VISARG arg;
	memset(&arg, 0, sizeof(arg));
	arg.svvisitor = svvisitor;
	arg.writable = writable;
	arg.rbuf = NULL;
	int32_t rv = kcdbaccept(db, kbuf, ksiz, softvisitfull, softvisitempty,
	                        &arg, writable);
	if (arg.rbuf) kcfree(arg.rbuf);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_iterate(IV ivdb, SV* svvisitor, SV* svwritable)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	int32_t writable = SvTRUE(svwritable);
	VISARG arg;
	memset(&arg, 0, sizeof(arg));
	arg.svvisitor = svvisitor;
	arg.writable = writable;
	arg.rbuf = NULL;
	int32_t rv = kcdbiterate(db, softvisitfull, &arg, writable);
	if (arg.rbuf) kcfree(arg.rbuf);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_close(IV ivdb)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	int32_t rv = kcdbclose(db);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_set(IV ivdb, SV* svkey, SV* svvalue)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t ksiz;
	const char* kbuf = SvPV(svkey, ksiz);
	size_t vsiz;
	const char* vbuf = SvPV(svvalue, vsiz);
	int32_t rv = kcdbset(db, kbuf, ksiz, vbuf, vsiz);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_add(IV ivdb, SV* svkey, SV* svvalue)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t ksiz;
	const char* kbuf = SvPV(svkey, ksiz);
	size_t vsiz;
	const char* vbuf = SvPV(svvalue, vsiz);
	int32_t rv = kcdbadd(db, kbuf, ksiz, vbuf, vsiz);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_replace(IV ivdb, SV* svkey, SV* svvalue)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t ksiz;
	const char* kbuf = SvPV(svkey, ksiz);
	size_t vsiz;
	const char* vbuf = SvPV(svvalue, vsiz);
	int32_t rv = kcdbreplace(db, kbuf, ksiz, vbuf, vsiz);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_append(IV ivdb, SV* svkey, SV* svvalue)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t ksiz;
	const char* kbuf = SvPV(svkey, ksiz);
	size_t vsiz;
	const char* vbuf = SvPV(svvalue, vsiz);
	int32_t rv = kcdbappend(db, kbuf, ksiz, vbuf, vsiz);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_increment(IV ivdb, SV* svkey, SV* svnum, SV* svorig)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t ksiz;
	const char* kbuf = SvPV(svkey, ksiz);
	int64_t num = svatoi(svnum);
	int64_t orig = svatoi(svorig);
	num = kcdbincrint(db, kbuf, ksiz, num, orig);
	if (num == INT64_MIN) {
	  XPUSHs(&PL_sv_undef);
	} else {
	  XPUSHs(sv_2mortal(newSViv(num)));
	}
	XSRETURN(1);


void
db_increment_double(IV ivdb, SV* svkey, SV* svnum, SV* svorig)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t ksiz;
	const char* kbuf = SvPV(svkey, ksiz);
	double num = svatof(svnum);
	double orig = svatof(svorig);
	num = kcdbincrdouble(db, kbuf, ksiz, num, orig);
	if (kcchknan(num)) {
	  XPUSHs(&PL_sv_undef);
	} else {
	  XPUSHs(sv_2mortal(newSVnv(num)));
	}
	XSRETURN(1);


void
db_cas(IV ivdb, SV* svkey, SV* svoval, SV* svnval)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t ksiz;
	const char* kbuf = SvPV(svkey, ksiz);
	const char* ovbuf = NULL;
	size_t ovsiz = 0;
	if (SvOK(svoval)) ovbuf = SvPV(svoval, ovsiz);
	const char* nvbuf = NULL;
	size_t nvsiz = 0;
	if (SvOK(svnval)) nvbuf = SvPV(svnval, nvsiz);
	int32_t rv = kcdbcas(db, kbuf, ksiz, ovbuf, ovsiz, nvbuf, nvsiz);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_remove(IV ivdb, SV* svkey)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t ksiz;
	const char* kbuf = SvPV(svkey, ksiz);
	int32_t rv = kcdbremove(db, kbuf, ksiz);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_get(IV ivdb, SV* svkey)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t ksiz;
	const char* kbuf = SvPV(svkey, ksiz);
	size_t vsiz;
	char* vbuf = kcdbget(db, kbuf, ksiz, &vsiz);
	if (vbuf) {
	  XPUSHs(sv_2mortal(newSVpvn(vbuf, vsiz)));
	  kcfree(vbuf);
	} else {
	  XPUSHs(&PL_sv_undef);
	}
	XSRETURN(1);


void
db_check(IV ivdb, SV* svkey)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t ksiz;
	const char* kbuf = SvPV(svkey, ksiz);
	int32_t vsiz = kcdbcheck(db, kbuf, ksiz);
	XPUSHs(sv_2mortal(newSViv(vsiz)));
	XSRETURN(1);


void
db_seize(IV ivdb, SV* svkey)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t ksiz;
	const char* kbuf = SvPV(svkey, ksiz);
	size_t vsiz;
	char* vbuf = kcdbseize(db, kbuf, ksiz, &vsiz);
	if (vbuf) {
	  XPUSHs(sv_2mortal(newSVpvn(vbuf, vsiz)));
	  kcfree(vbuf);
	} else {
	  XPUSHs(&PL_sv_undef);
	}
	XSRETURN(1);


void
db_clear(IV ivdb)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	int32_t rv = kcdbclear(db);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_synchronize(IV ivdb, SV* svhard, SV* svproc)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	int32_t hard = SvTRUE(svhard);
	int32_t rv = kcdbsync(db, hard, softfileproc, svproc);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_occupy(IV ivdb, SV* svwritable, SV* svproc)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	int32_t writable = SvTRUE(svwritable);
	int32_t rv = kcdboccupy(db, writable, softfileproc, svproc);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_copy(IV ivdb, SV* svdest)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t dsiz;
	const char* dest = SvPV(svdest, dsiz);
	int32_t rv = kcdbcopy(db, dest);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_begin_transaction(IV ivdb, SV* svhard)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	int32_t hard = SvTRUE(svhard);
	int32_t rv = kcdbbegintran(db, hard);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_end_transaction(IV ivdb, SV* svcommit)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	int32_t commit = SvTRUE(svcommit);
	int32_t rv = kcdbendtran(db, commit);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_dump_snapshot(IV ivdb, SV* svdest)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t dsiz;
	const char* dest = SvPV(svdest, dsiz);
	int32_t rv = kcdbdumpsnap(db, dest);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_load_snapshot(IV ivdb, SV* svsrc)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t ssiz;
	const char* src = SvPV(svsrc, ssiz);
	int32_t rv = kcdbloadsnap(db, src);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;


void
db_count(IV ivdb)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	int64_t count = kcdbcount(db);
	XPUSHs(sv_2mortal(newSViv(count)));
	XSRETURN(1);


void
db_size(IV ivdb)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	int64_t size = kcdbsize(db);
	XPUSHs(sv_2mortal(newSViv(size)));
	XSRETURN(1);


void
db_path(IV ivdb)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	char* path = kcdbpath(db);
	if (*path == '\0') {
	  XPUSHs(&PL_sv_undef);
	} else {
	  XPUSHs(sv_2mortal(newSVpv(path, 0)));
	}
	kcfree(path);
	XSRETURN(1);


void
db_status(IV ivdb)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	char* status = kcdbstatus(db);
	if (status) {
	  XPUSHs(sv_2mortal(newSVpv(status, 0)));
	  kcfree(status);
	} else {
	  XPUSHs(&PL_sv_undef);
	}
	XSRETURN(1);


void
db_match_prefix(IV ivdb, SV* svprefix, SV* svmax)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t psiz;
	const char* pbuf = SvPV(svprefix, psiz);
	int64_t max = svatoi(svmax);
	if (max < 0) max = 1LL << 20 - 1;
	if (max >= 65536) {
	  int64_t cnt = kcdbcount(db);
	  if (cnt >= 0 && max > cnt) max = cnt;
	}
	char** strary = kcmalloc(sizeof(*strary) * max + 1);
	int64_t num = kcdbmatchprefix(db, pbuf, strary, max);
	if (num >= 0) {
	  AV* av = newAV();
	  int64_t i;
	  for (i = 0; i < num; i++) {
	    if (i < max) av_push(av, newSVpv(strary[i], 0));
	    kcfree(strary[i]);
	  }
	  XPUSHs(sv_2mortal(newRV_noinc((SV*)av)));
	} else {
	  XPUSHs(&PL_sv_undef);
	}
	kcfree(strary);
	XSRETURN(1);


void
db_match_regex(IV ivdb, SV* svregex, SV* svmax)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t rsiz;
	const char* rbuf = SvPV(svregex, rsiz);
	int64_t max = svatoi(svmax);
	if (max < 0) max = 1LL << 20 - 1;
	if (max >= 65536) {
	  int64_t cnt = kcdbcount(db);
	  if (cnt >= 0 && max > cnt) max = cnt;
	}
	char** strary = kcmalloc(sizeof(*strary) * max + 1);
	int64_t num = kcdbmatchregex(db, rbuf, strary, max);
	if (num >= 0) {
	  AV* av = newAV();
	  int64_t i;
	  for (i = 0; i < num; i++) {
	    if (i < max) av_push(av, newSVpv(strary[i], 0));
	    kcfree(strary[i]);
	  }
	  XPUSHs(sv_2mortal(newRV_noinc((SV*)av)));
	} else {
	  XPUSHs(&PL_sv_undef);
	}
	kcfree(strary);
	XSRETURN(1);


void
db_match_similar(IV ivdb, SV* svorigin, SV* svrange, SV* svutf, SV* svmax)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	size_t osiz;
	const char* obuf = SvPV(svorigin, osiz);
	int64_t range = svatoi(svrange);
	int32_t utf = SvTRUE(svutf);
	int64_t max = svatoi(svmax);
	if (max < 0) max = 1LL << 20 - 1;
	if (max >= 65536) {
	  int64_t cnt = kcdbcount(db);
	  if (cnt >= 0 && max > cnt) max = cnt;
	}
	char** strary = kcmalloc(sizeof(*strary) * max + 1);
	int64_t num = kcdbmatchsimilar(db, obuf, range, utf, strary, max);
	if (num >= 0) {
	  AV* av = newAV();
	  int64_t i;
	  for (i = 0; i < num; i++) {
	    if (i < max) av_push(av, newSVpv(strary[i], 0));
	    kcfree(strary[i]);
	  }
	  XPUSHs(sv_2mortal(newRV_noinc((SV*)av)));
	} else {
	  XPUSHs(&PL_sv_undef);
	}
	kcfree(strary);
	XSRETURN(1);


void
db_merge(IV ivdb, SV* svsrcary, IV mode)
PPCODE:
	KCDB* db = (KCDB*)ivdb;
	if (!SvROK(svsrcary) || SvTYPE(SvRV(svsrcary)) != SVt_PVAV) XSRETURN_UNDEF;
	AV* avsrcary = (AV*)SvRV(svsrcary);
	int32_t num = av_len(avsrcary);
	if (num < 0) XSRETURN_YES;
	KCDB** srcary = (KCDB**)kcmalloc(sizeof(*srcary) * num);
	size_t srcnum = 0;
	int32_t i;
	for (i = 0; i <= num; i++) {
	  SV* svsrcdb = *av_fetch(avsrcary, i, 0);
	  if (SvROK(svsrcdb) && SvTYPE(SvRV(svsrcdb)) == SVt_PVAV) {
	    AV* avsrcdb = (AV*)SvRV(svsrcdb);
	    if (av_len(avsrcdb) >= 0) {
	      svsrcdb = *av_fetch(avsrcdb, 0, 0);
	      srcary[srcnum++] = (KCDB*)SvIV(svsrcdb);
	    }
	  }
	}
	int32_t rv = kcdbmerge(db, srcary, srcnum, mode);
	kcfree((void*)srcary);
	if (!rv) XSRETURN_NO;
	XSRETURN_YES;



## END OF FILE
